"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionInputDefaultFragment = void 0;
const shared_1 = require("../../../shared");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
const valueNode_1 = require("./valueNode");
function getInstructionInputDefaultFragment(scope) {
    const { input, optionalAccountStrategy, asyncResolvers, useAsync, nameApi } = scope;
    if (!input.defaultsTo) {
        return (0, common_1.fragment)('');
    }
    if (!useAsync && (0, asyncHelpers_1.isAsyncDefaultValue)(input.defaultsTo, asyncResolvers)) {
        return (0, common_1.fragment)('');
    }
    const { defaultsTo } = input;
    const defaultFragment = (defaultValue, isWritable) => {
        const inputName = (0, shared_1.camelCase)(input.name);
        if (input.kind === 'account' && defaultsTo.kind === 'resolver') {
            return (0, common_1.fragment)(`accounts.${inputName} = { ...accounts.${inputName}, ...${defaultValue} };`);
        }
        if (input.kind === 'account' && isWritable === undefined) {
            return (0, common_1.fragment)(`accounts.${inputName}.value = ${defaultValue};`);
        }
        if (input.kind === 'account') {
            return (0, common_1.fragment)(`accounts.${inputName}.value = ${defaultValue};\n` +
                `accounts.${inputName}.isWritable = ${isWritable ? 'true' : 'false'}`);
        }
        return (0, common_1.fragment)(`args.${inputName} = ${defaultValue};`);
    };
    switch (defaultsTo.kind) {
        case 'account':
            const name = (0, shared_1.camelCase)(defaultsTo.name);
            if (input.kind === 'account' &&
                input.resolvedIsSigner &&
                !input.isSigner) {
                return defaultFragment(`expectTransactionSigner(accounts.${name}.value).address`).addImports('shared', 'expectTransactionSigner');
            }
            if (input.kind === 'account') {
                return defaultFragment(`expectSome(accounts.${name}.value)`).addImports('shared', 'expectSome');
            }
            return defaultFragment(`expectAddress(accounts.${name}.value)`).addImports('shared', 'expectAddress');
        case 'pda':
            const pdaFunction = nameApi.accountFindPdaFunction(defaultsTo.pdaAccount);
            const pdaImportFrom = defaultsTo.importFrom === 'generated'
                ? 'generatedAccounts'
                : defaultsTo.importFrom;
            const pdaArgs = [];
            const pdaSeeds = Object.keys(defaultsTo.seeds).map((seed) => {
                const seedValue = defaultsTo.seeds[seed];
                if (seedValue.kind === 'account') {
                    return (0, common_1.fragment)(`${seed}: expectAddress(accounts.${(0, shared_1.camelCase)(seedValue.name)}.value)`).addImports('shared', 'expectAddress');
                }
                if (seedValue.kind === 'arg') {
                    return (0, common_1.fragment)(`${seed}: expectSome(args.${(0, shared_1.camelCase)(seedValue.name)})`).addImports('shared', 'expectSome');
                }
                return (0, valueNode_1.getValueNodeFragment)(seedValue.value, nameApi).mapRender((r) => `${seed}: ${r}`);
            });
            const pdaSeedsFragment = (0, common_1.mergeFragments)(pdaSeeds, (renders) => renders.join(', ')).mapRender((r) => `{ ${r} }`);
            if (pdaSeeds.length > 0) {
                pdaArgs.push(pdaSeedsFragment.render);
            }
            return defaultFragment(`await ${pdaFunction}(${pdaArgs.join(', ')})`)
                .mergeImportsWith(pdaSeedsFragment)
                .addImports(pdaImportFrom, pdaFunction);
        case 'publicKey':
            return defaultFragment(`'${defaultsTo.publicKey}' as Address<'${defaultsTo.publicKey}'>`).addImports('solanaAddresses', 'Address');
        case 'program':
            return defaultFragment(`'${defaultsTo.program.publicKey}' as Address<'${defaultsTo.program.publicKey}'>`, false).addImports('solanaAddresses', ['Address']);
        case 'programId':
            if (optionalAccountStrategy === 'programId' &&
                input.kind === 'account' &&
                input.isOptional) {
                return (0, common_1.fragment)('');
            }
            return defaultFragment('programAddress', false);
        // Deprecated
        case 'identity':
        case 'payer':
            return (0, common_1.fragment)('');
        case 'accountBump':
            return defaultFragment(`expectProgramDerivedAddress(accounts.${(0, shared_1.camelCase)(defaultsTo.name)}.value)[1]`).addImports('shared', 'expectProgramDerivedAddress');
        case 'arg':
            return defaultFragment(`expectSome(args.${(0, shared_1.camelCase)(defaultsTo.name)})`).addImports('shared', 'expectSome');
        case 'value':
            const valueManifest = (0, valueNode_1.getValueNodeFragment)(defaultsTo.value, nameApi);
            return defaultFragment(valueManifest.render).mergeImportsWith(valueManifest);
        case 'resolver':
            const resolverFunction = nameApi.resolverFunction(defaultsTo.name);
            const resolverAwait = useAsync && asyncResolvers.includes(defaultsTo.name) ? 'await ' : '';
            return defaultFragment(`${resolverAwait}${resolverFunction}(resolverScope)`)
                .addImports(defaultsTo.importFrom, resolverFunction)
                .addFeatures(['instruction:resolverScopeVariable']);
        case 'conditional':
        case 'conditionalResolver':
            const ifTrueRenderer = renderNestedInstructionDefault({
                ...scope,
                defaultsTo: defaultsTo.ifTrue,
            });
            const ifFalseRenderer = renderNestedInstructionDefault({
                ...scope,
                defaultsTo: defaultsTo.ifFalse,
            });
            if (!ifTrueRenderer && !ifFalseRenderer) {
                return (0, common_1.fragment)('');
            }
            const conditionalFragment = (0, common_1.fragment)('');
            if (ifTrueRenderer) {
                conditionalFragment
                    .mergeImportsWith(ifTrueRenderer)
                    .mergeFeaturesWith(ifTrueRenderer);
            }
            if (ifFalseRenderer) {
                conditionalFragment
                    .mergeImportsWith(ifFalseRenderer)
                    .mergeFeaturesWith(ifFalseRenderer);
            }
            const negatedCondition = !ifTrueRenderer;
            let condition = 'true';
            if (defaultsTo.kind === 'conditional') {
                const comparedInputName = defaultsTo.input.kind === 'account'
                    ? `accounts.${(0, shared_1.camelCase)(defaultsTo.input.name)}.value`
                    : `args.${(0, shared_1.camelCase)(defaultsTo.input.name)}`;
                if (defaultsTo.value) {
                    const comparedValue = (0, valueNode_1.getValueNodeFragment)(defaultsTo.value, nameApi);
                    conditionalFragment
                        .mergeImportsWith(comparedValue)
                        .mergeFeaturesWith(comparedValue);
                    const operator = negatedCondition ? '!==' : '===';
                    condition = `${comparedInputName} ${operator} ${comparedValue.render}`;
                }
                else {
                    condition = negatedCondition
                        ? `!${comparedInputName}`
                        : comparedInputName;
                }
            }
            else {
                const conditionalResolverFunction = nameApi.resolverFunction(defaultsTo.resolver.name);
                conditionalFragment
                    .addImports(defaultsTo.resolver.importFrom, conditionalResolverFunction)
                    .addFeatures(['instruction:resolverScopeVariable']);
                const conditionalResolverAwait = useAsync && asyncResolvers.includes(defaultsTo.resolver.name)
                    ? 'await '
                    : '';
                condition = `${conditionalResolverAwait}${conditionalResolverFunction}(resolverScope)`;
                condition = negatedCondition ? `!${condition}` : condition;
            }
            if (ifTrueRenderer && ifFalseRenderer) {
                return conditionalFragment.setRender(`if (${condition}) {\n${ifTrueRenderer.render}\n} else {\n${ifFalseRenderer.render}\n}`);
            }
            return conditionalFragment.setRender(`if (${condition}) {\n${ifTrueRenderer ? ifTrueRenderer.render : ifFalseRenderer === null || ifFalseRenderer === void 0 ? void 0 : ifFalseRenderer.render}\n}`);
        default:
            const neverDefault = defaultsTo;
            throw new Error(`Unexpected value type ${neverDefault.kind}`);
    }
}
exports.getInstructionInputDefaultFragment = getInstructionInputDefaultFragment;
function renderNestedInstructionDefault(scope) {
    const { input, defaultsTo } = scope;
    if (!defaultsTo)
        return undefined;
    if (input.kind === 'account') {
        return getInstructionInputDefaultFragment({
            ...scope,
            input: { ...input, defaultsTo: defaultsTo },
        });
    }
    return getInstructionInputDefaultFragment({
        ...scope,
        input: { ...input, defaultsTo: defaultsTo },
    });
}
//# sourceMappingURL=instructionInputDefault.js.map