"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionInputTypeFragment = void 0;
const shared_1 = require("../../../shared");
const ImportMap_1 = require("../ImportMap");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
function getInstructionInputTypeFragment(scope) {
    const { instructionNode, resolvedInputs, renamedArgs, dataArgsManifest, extraArgsManifest, programNode, withSigners, asyncResolvers, useAsync, nameApi, } = scope;
    // Accounts.
    const accountImports = new ImportMap_1.ImportMap();
    const accounts = instructionNode.accounts.map((account) => {
        const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
        const resolvedAccount = resolvedInputs.find((input) => input.kind === 'account' && input.name === account.name);
        const hasDefaultValue = !!resolvedAccount.defaultsTo &&
            !['identity', 'payer'].includes(resolvedAccount.defaultsTo.kind) &&
            (useAsync ||
                !(0, asyncHelpers_1.isAsyncDefaultValue)(resolvedAccount.defaultsTo, asyncResolvers));
        const type = getAccountType(resolvedAccount, withSigners);
        accountImports.mergeWith(type);
        return {
            ...resolvedAccount,
            typeParam,
            optionalSign: hasDefaultValue || resolvedAccount.isOptional ? '?' : '',
            type: type.render,
        };
    });
    // Arg link imports.
    const argLinkImports = new ImportMap_1.ImportMap();
    if (instructionNode.dataArgs.link) {
        argLinkImports.mergeWith(dataArgsManifest.looseType);
    }
    if (instructionNode.extraArgs.link) {
        argLinkImports.mergeWith(extraArgsManifest.looseType);
    }
    // Arguments.
    const resolveArg = (arg) => {
        var _a, _b;
        const resolvedArg = resolvedInputs.find((input) => input.kind === 'arg' && input.name === arg.name);
        if (((_a = arg.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) === 'omitted')
            return [];
        const renamedName = (_b = renamedArgs.get(arg.name)) !== null && _b !== void 0 ? _b : arg.name;
        const optionalSign = arg.defaultsTo || resolvedArg ? '?' : '';
        return [
            {
                ...arg,
                ...resolvedArg,
                renamedName,
                optionalSign,
            },
        ];
    };
    const dataArgsType = instructionNode.dataArgs.link
        ? nameApi.dataArgsType(instructionNode.dataArgs.link.name)
        : nameApi.dataArgsType(instructionNode.dataArgs.name);
    const dataArgs = instructionNode.dataArgs.link
        ? []
        : instructionNode.dataArgs.struct.fields.flatMap(resolveArg);
    const extraArgsType = instructionNode.extraArgs.link
        ? nameApi.dataArgsType(instructionNode.extraArgs.link.name)
        : nameApi.dataArgsType(instructionNode.extraArgs.name);
    const extraArgs = instructionNode.extraArgs.link
        ? []
        : instructionNode.extraArgs.struct.fields.flatMap(resolveArg);
    const syncInputType = withSigners
        ? nameApi.instructionSyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionSyncInputType(instructionNode.name);
    const asyncInputType = withSigners
        ? nameApi.instructionAsyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionAsyncInputType(instructionNode.name);
    const instructionInputType = useAsync ? asyncInputType : syncInputType;
    return (0, common_1.fragmentFromTemplate)('instructionInputType.njk', {
        instruction: instructionNode,
        program: programNode,
        instructionInputType,
        accounts,
        dataArgs,
        dataArgsType,
        extraArgs,
        extraArgsType,
    })
        .mergeImportsWith(accountImports, argLinkImports)
        .addImports('solanaAddresses', ['Address']);
}
exports.getInstructionInputTypeFragment = getInstructionInputTypeFragment;
function getAccountType(account, withSigners) {
    const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
    if (withSigners) {
        if (account.isPda && account.isSigner === false) {
            return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}>`).addImports('solanaAddresses', ['ProgramDerivedAddress']);
        }
        if (account.isPda && account.isSigner === 'either') {
            return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}> | TransactionSigner<${typeParam}>`)
                .addImports('solanaAddresses', ['ProgramDerivedAddress'])
                .addImports('solanaSigners', ['TransactionSigner']);
        }
        if (account.isSigner === 'either') {
            return (0, common_1.fragment)(`Address<${typeParam}> | TransactionSigner<${typeParam}>`)
                .addImports('solanaAddresses', ['Address'])
                .addImports('solanaSigners', ['TransactionSigner']);
        }
        if (account.isSigner) {
            return (0, common_1.fragment)(`TransactionSigner<${typeParam}>`).addImports('solanaSigners', ['TransactionSigner']);
        }
    }
    if (!withSigners && account.isPda) {
        return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}>`).addImports('solanaAddresses', ['ProgramDerivedAddress']);
    }
    return (0, common_1.fragment)(`Address<${typeParam}>`).addImports('solanaAddresses', [
        'Address',
    ]);
}
//# sourceMappingURL=instructionInputType.js.map